from threading import Thread


class ThreadWithReturnValue(Thread):
    """
    A custom Thread class that allows the target function to return a value.
    This class extends the built-in threading.Thread class.
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, verbose=None):
        """
        Initializes the ThreadWithReturnValue object.

        Parameters:
        group (threading.ThreadGroup)
        target (callable): The callable object to be invoked by the run() method. Defaults to None.
        name (str): The thread name. Defaults to None.
        args (tuple): The argument tuple for the target invocation. Defaults to ().
        kwargs (dict): The dictionary of keyword arguments for the target invocation.
         Defaults to {}.
        verbose: Not used, defaults to None.
        """
        Thread.__init__(self, group, target, name, args, kwargs)
        # Variable to store the return value of the target function
        self._return = None

    def run(self):
        """
        Run the thread.

        This method is overridden from the Thread class.
        It calls the target function and saves the return value.
        """
        if self._target is not None:
            # Execute target and store the result
            self._return = self._target(*self._args, **self._kwargs)

    def join(self, *args):
        """
        Wait for the thread to complete and return the result.

        This method is overridden from the Thread class.
        It calls the parent class's join() method and then returns the stored return value.

        Parameters:
        *args: Variable length argument list to pass to the join() method.

        Returns:
        The return value from the target function.
        """
        # Call the parent class's join() method to wait for the thread to finish
        Thread.join(self, *args)
        # Return the result of the target function
        return self._return



from dataclasses import dataclass, asdict


@dataclass
class _RuntimeErrorEntry:
    """Represents a runtime error entry to be persisted."""

    container: str
    timestamp: int
    code: str
    message: str
    details: str

    def to_dict(self) -> dict:
        """Converts the dataclass instance to a dictionary."""
        return asdict(self)



from pynumaflow.info.types import (
    ServerInfo,
    MAP_MODE_KEY,
    MapMode,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.mapper._servicer._sync_servicer import SyncMapServicer

from pynumaflow._constants import (
    NUM_THREADS_DEFAULT,
    MAX_MESSAGE_SIZE,
    _LOGGER,
    MAP_SOCK_PATH,
    UDFType,
    MAP_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.mapper._dtypes import MapSyncCallable
from pynumaflow.shared.server import (
    NumaflowServer,
    sync_server_start,
)


class MapServer(NumaflowServer):
    """
    Create a new grpc Map Server instance.
    Args:
        mapper_instance: The mapper instance to be used for Map UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16

    Example Invocation:
        from pynumaflow.mapper import Messages, Message, Datum, MapServer, Mapper

        class MessageForwarder(Mapper):
            def handler(self, keys: list[str], datum: Datum) -> Messages:
                val = datum.value
                _ = datum.event_time
                _ = datum.watermark
                return Messages(Message(value=val, keys=keys))

        def my_handler(keys: list[str], datum: Datum) -> Messages:
            val = datum.value
            _ = datum.event_time
            _ = datum.watermark
            return Messages(Message(value=val, keys=keys))


        if __name__ == "__main__":
            Use the class based approach or function based handler
            based on the env variable
            Both can be used and passed directly to the server class

            invoke = os.getenv("INVOKE", "func_handler")
            if invoke == "class":
                handler = MessageForwarder()
            else:
                handler = my_handler
            grpc_server = MapServer(handler)
            grpc_server.start()
    """

    def __init__(
        self,
        mapper_instance: MapSyncCallable,
        sock_path=MAP_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=MAP_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Synchronous Map Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
        mapper_instance: The mapper instance to be used for Map UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to 4 and max capped at 16
        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.mapper_instance = mapper_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        # Get the servicer instance for the sync server
        self.servicer = SyncMapServicer(handler=mapper_instance)

    def start(self) -> None:
        """
        Starts the Synchronous gRPC server on the given UNIX socket with given max threads.
        """
        _LOGGER.info(
            "Sync GRPC Server listening on: %s with max threads: %s",
            self.sock_path,
            self.max_threads,
        )

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Mapper]
        # Add the MAP_MODE metadata to the server info for the correct map mode
        serv_info.metadata[MAP_MODE_KEY] = MapMode.UnaryMap
        # Start the server
        sync_server_start(
            servicer=self.servicer,
            bind_address=self.sock_path,
            max_threads=self.max_threads,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.Map,
            server_info=serv_info,
        )



import aiorun
import grpc

from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    BATCH_MAP_SOCK_PATH,
    MAP_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)
from pynumaflow.batchmapper._dtypes import BatchMapCallable
from pynumaflow.batchmapper.servicer.async_servicer import AsyncBatchMapServicer
from pynumaflow.info.types import (
    ServerInfo,
    MAP_MODE_KEY,
    MapMode,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.proto.mapper import map_pb2_grpc
from pynumaflow.shared.server import NumaflowServer, start_async_server


class BatchMapAsyncServer(NumaflowServer):
    """
    Class for a new Batch Map Async Server instance.
    """

    def __init__(
        self,
        batch_mapper_instance: BatchMapCallable,
        sock_path=BATCH_MAP_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=MAP_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Async Batch Map Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            batch_mapper_instance: The batch map stream instance to be used for Batch Map UDF
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16

        Example invocation:
         class Flatmap(BatchMapper):
            async def handler(
                self,
                datums: AsyncIterable[Datum],
            ) -> BatchResponses:
                batch_responses = BatchResponses()
                async for datum in datums:
                    val = datum.value
                    _ = datum.event_time
                    _ = datum.watermark
                    strs = val.decode("utf-8").split(",")
                    batch_response = BatchResponse.from_id(datum.id)
                    if len(strs) == 0:
                        batch_response.append(Message.to_drop())
                    else:
                        for s in strs:
                            batch_response.append(Message(str.encode(s)))
                    batch_responses.append(batch_response)

                return batch_responses

        if __name__ == "__main__":
            grpc_server = BatchMapAsyncServer(Flatmap())
            grpc_server.start()
        """
        self.batch_mapper_instance: BatchMapCallable = batch_mapper_instance
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]

        self.servicer = AsyncBatchMapServicer(handler=self.batch_mapper_instance)

    def start(self):
        """
        Starter function for the Async Batch Map server, we need a separate caller
        to the aexec so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        map_pb2_grpc.add_MapServicer_to_server(
            self.servicer,
            server,
        )
        _LOGGER.info("Starting Batch Map Server")
        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Mapper]
        # Add the MAP_MODE metadata to the server info for the correct map mode
        serv_info.metadata[MAP_MODE_KEY] = MapMode.BatchMap

        # Start the async server
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



class NoPublicConstructorError(TypeError):
    """Raise when using ClassName() to create objects while public constructor is not supported"""


class SocketError(Exception):
    """To raise an error while creating socket or setting its property"""


class UDFError(Exception):
    """To Raise an error while executing a UDF call"""



import logging
import os
from enum import Enum

from pynumaflow import setup_logging

SIDE_INPUT_DIR_PATH = "/var/numaflow/side-inputs"
ENV_UD_CONTAINER_TYPE = "NUMAFLOW_UD_CONTAINER_TYPE"

# Error Constants
RUNTIME_APPLICATION_ERRORS_PATH = "/var/numaflow/runtime/application-errors"
CURRENT_CRITICAL_ERROR_FILE = "current-udf.json"
INTERNAL_ERROR_CODE = "Internal error"
CONTAINER_TYPE = os.getenv(ENV_UD_CONTAINER_TYPE, "unknown-container")
ERR_UDF_EXCEPTION_STRING = f"UDF_EXECUTION_ERROR({CONTAINER_TYPE})"

# Socket configs
MAP_SOCK_PATH = "/var/run/numaflow/map.sock"
MAP_STREAM_SOCK_PATH = "/var/run/numaflow/mapstream.sock"
REDUCE_SOCK_PATH = "/var/run/numaflow/reduce.sock"
REDUCE_STREAM_SOCK_PATH = "/var/run/numaflow/reducestream.sock"
SOURCE_TRANSFORMER_SOCK_PATH = "/var/run/numaflow/sourcetransform.sock"
SINK_SOCK_PATH = "/var/run/numaflow/sink.sock"
SIDE_INPUT_SOCK_PATH = "/var/run/numaflow/sideinput.sock"
SOURCE_SOCK_PATH = "/var/run/numaflow/source.sock"
MULTIPROC_MAP_SOCK_ADDR = "/var/run/numaflow/multiproc"
FALLBACK_SINK_SOCK_PATH = "/var/run/numaflow/fb-sink.sock"
BATCH_MAP_SOCK_PATH = "/var/run/numaflow/batchmap.sock"
ACCUMULATOR_SOCK_PATH = "/var/run/numaflow/accumulator.sock"

# Server information file configs
MAP_SERVER_INFO_FILE_PATH = "/var/run/numaflow/mapper-server-info"
REDUCE_SERVER_INFO_FILE_PATH = "/var/run/numaflow/reducer-server-info"
REDUCE_STREAM_SERVER_INFO_FILE_PATH = "/var/run/numaflow/reducestreamer-server-info"
SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH = "/var/run/numaflow/sourcetransformer-server-info"
SINK_SERVER_INFO_FILE_PATH = "/var/run/numaflow/sinker-server-info"
SIDE_INPUT_SERVER_INFO_FILE_PATH = "/var/run/numaflow/sideinput-server-info"
SOURCE_SERVER_INFO_FILE_PATH = "/var/run/numaflow/sourcer-server-info"
FALLBACK_SINK_SERVER_INFO_FILE_PATH = "/var/run/numaflow/fb-sinker-server-info"
ACCUMULATOR_SERVER_INFO_FILE_PATH = "/var/run/numaflow/accumulator-server-info"

ENV_UD_CONTAINER_TYPE = "NUMAFLOW_UD_CONTAINER_TYPE"
UD_CONTAINER_FALLBACK_SINK = "fb-udsink"

# TODO: need to make sure the DATUM_KEY value is the same as
# https://github.com/numaproj/numaflow-go/blob/main/pkg/function/configs.go#L6
WIN_START_TIME = "x-numaflow-win-start-time"
WIN_END_TIME = "x-numaflow-win-end-time"
MAX_MESSAGE_SIZE = 1024 * 1024 * 64
# TODO: None instead of "EOF" ?
STREAM_EOF = "EOF"
DELIMITER = ":"
DROP = "U+005C__DROP__"

_PROCESS_COUNT = os.cpu_count()
# Cap max value to 16
MAX_NUM_THREADS = 16
# If NUM_THREADS_DEFAULT env is not set default to 4
NUM_THREADS_DEFAULT = int(os.getenv("MAX_THREADS", "4"))

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)


class UDFType(str, Enum):
    """
    Enumerate the type of UDF.
    """

    Map = "map"
    Reduce = "reduce"
    Sink = "sink"
    Source = "source"
    SideInput = "sideinput"
    SourceTransformer = "sourcetransformer"
